class CaptionReadError(Exception):
    """
    Generic error raised when the reading of the caption file failed.
    """

    def __str__(self):
        return f"{self.__class__.__name__}({self.args[0]})"


class CaptionReadNoCaptions(CaptionReadError):
    """
    Error raised when the provided caption file was not containing any
    actual captions.
    """


class CaptionReadSyntaxError(CaptionReadError):
    """
    Error raised when the provided caption file has syntax errors and could
    not be parsed.
    """


class CaptionReadTimingError(CaptionReadError):
    """
    Error raised when a Caption is initialized with invalid timings.
    """


class RelativizationError(Exception):
    """
    Error raised when absolute positioning cannot be converted to
    percentage
    """


class InvalidInputError(RuntimeError):
    """Error raised when the input is invalid (i.e. a unicode string)"""


class CaptionLineLengthError(CaptionReadError):
    """
    Error raised when a Caption has a line longer than 32 characters.
    """



from ..exceptions import CaptionReadSyntaxError


class _PositioningTracker:
    """Helps determine the positioning of a node, having kept track of
    positioning-related commands.
    """

    def __init__(self, positioning=None):
        """
        :param positioning: positioning information (row, column)
        :type positioning: tuple[int]
        """
        self._positions = [positioning]
        self._breaks_required = 0
        self._repositioning_required = False
        # Since the actual column is not applied when encountering a line break
        # this attribute is used to store it and determine by comparison if the
        # next positioning is actually a Tab Offset
        self._last_column = None

    def update_positioning(self, positioning):
        """Being notified of a position change, updates the internal state,
        to as to be able to tell if it was a trivial change (a simple line
        break) or not.

        Strategy:
        - Small jumps (1-3 rows): Use line breaks to preserve visual spacing
        - Large jumps (4+ rows): Use repositioning (creates new cue)

        :type positioning: tuple[int]
        :param positioning: a tuple (row, col)
        """
        current = self._positions[-1]

        if not current:
            if positioning:
                # Set the positioning for the first time
                self._positions = [positioning]
            return

        row, col = current
        if self._breaks_required:
            col = self._last_column
        new_row, new_col = positioning
        is_tab_offset = new_row == row and col + 1 <= new_col <= col + 3

        # Threshold for when to use breaks vs repositioning
        # Jumps of 4+ rows will trigger repositioning instead of adding breaks
        MAX_BREAKS_THRESHOLD = 3

        # Handle row jumps
        if new_row > row:
            row_diff = new_row - row

            # Small jumps (1-3 rows): Use line breaks to preserve visual spacing
            if row_diff <= MAX_BREAKS_THRESHOLD:
                self._positions.append((new_row, col))
                # Add breaks equal to row difference
                # Row N -> N+1: 1 break
                # Row N -> N+2: 2 breaks (preserves 1 blank line)
                # Row N -> N+3: 3 breaks (preserves 2 blank lines)
                self._breaks_required = row_diff
                self._last_column = new_col
            # Large jumps (4+ rows): Use repositioning (new cue)
            else:
                # Reset position - this triggers repositioning
                self._positions = [positioning]
                self._repositioning_required = True
        # Tab offsets after line breaks will be ignored to avoid repositioning
        elif self._breaks_required and is_tab_offset:
            return
        # force not to reposition on the same coordinates
        elif positioning == current:
            return
        else:
            # Reset the "current" position altogether.
            self._positions = [positioning]
            # Tab offsets are not interpreted as repositioning, but adjustments
            # to the previous PAC command
            if not is_tab_offset:
                self._repositioning_required = True

    def get_current_position(self):
        """Returns the current usable position

        :rtype: tuple[int]

        :raise: CaptionReadSyntaxError
        """
        if not any(self._positions):
            raise CaptionReadSyntaxError("No Preamble Address Code [PAC] was provided")
        else:
            return self._positions[0]

    def is_repositioning_required(self):
        """Determines whether the current positioning has changed non-trivially

        Trivial would be mean that a line break should suffice.
        :rtype: bool
        """
        return self._repositioning_required

    def acknowledge_position_changed(self):
        """Acknowledge the position tracer that the position was changed"""
        self._repositioning_required = False

    def is_linebreak_required(self):
        """If the current position is simply one line below the previous.
        :rtype: bool
        """
        return self._breaks_required > 0

    def acknowledge_linebreak_consumed(self):
        """Call to acknowledge that the line required was consumed"""
        self._breaks_required = 0

    def reset_for_new_caption(self):
        """Reset positioning state for a new caption boundary (e.g., EOC).

        This ensures that breaks and repositioning state from the previous
        caption do not bleed into the new caption. The position list is reset
        to allow the next caption to set its position independently.
        """
        self._breaks_required = 0
        self._repositioning_required = False
        self._last_column = None
        # Reset positions to None so the next PAC sets position fresh
        self._positions = [None]


class DefaultProvidingPositionTracker(_PositioningTracker):
    """A _PositioningTracker that provides if needed a default value (14, 0), or
    uses the last positioning value set anywhere in the document
    """

    default = (14, 0)

    def __init__(self, positioning=None, default=None):
        """
        :type positioning: tuple[int]
        :param positioning: a tuple of ints (row, column)

        :type default: tuple[int]
        :param default: a tuple of ints (row, column) to use as fallback
        """
        super().__init__(positioning)

        if default:
            self.default = default

    def get_current_position(self):
        """Returns the currently tracked positioning, the last positioning that
        was set (anywhere), or the default it was initiated with

        :rtype: tuple[int]
        """
        try:
            return super().get_current_position()
        except CaptionReadSyntaxError:
            return self.default

    def update_positioning(self, positioning):
        """If called, sets this positioning as the default, then delegates
        to the super class.

        :param positioning: a tuple of ints (row, col)
        :type positioning: tuple[int]
        """
        if positioning:
            self.default = positioning

        super().update_positioning(positioning)



import re
from copy import deepcopy

from .base import (
    DEFAULT_LANGUAGE_CODE,
    BaseReader,
    BaseWriter,
    Caption,
    CaptionList,
    CaptionNode,
    CaptionSet,
)
from .exceptions import (
    CaptionReadNoCaptions,
    CaptionReadSyntaxError,
    CaptionReadTimingError,
    InvalidInputError,
)


class MicroDVDReader(BaseReader):
    def detect(self, content):
        return re.match(r"{\d+}{\d+}", content) is not None

    def read(self, content, lang=DEFAULT_LANGUAGE_CODE):
        if not isinstance(content, str):
            raise InvalidInputError("The content is not a unicode string.")

        lines = content.splitlines()
        captions = CaptionList()
        fps = 25.0
        for line in lines:
            if not line:
                continue

            m = re.match(r"{(\d+)}{(\d+)}(.*)", line)
            if not m:
                raise CaptionReadSyntaxError("Line does not match expected format")

            start, end, txt = m.groups()

            if start == "0" and end == "0":
                try:
                    fps = float(txt)
                    continue
                except ValueError:
                    raise CaptionReadTimingError("FPS information is not provided")

            caption_start = self._framestomicro(int(start), fps)
            caption_end = self._framestomicro(int(end), fps)
            nodes = []

            for line in txt.split("|"):
                # skip extra blank lines
                if line != "":
                    nodes.append(CaptionNode.create_text(line))
                    nodes.append(CaptionNode.create_break())

            # remove last line break from end of caption list
            if len(nodes):
                nodes.pop()

                caption = Caption(caption_start, caption_end, nodes)
                captions.append(caption)

        caption_set = CaptionSet({lang: captions})
        caption_set.set_captions(lang, captions)

        if caption_set.is_empty():
            raise CaptionReadNoCaptions("Empty caption file")

        return caption_set

    def _framestomicro(self, framenum, fps=25.0):
        return int(framenum / fps * (10**6))


class MicroDVDWriter(BaseWriter):
    def write(self, caption_set):
        caption_set = deepcopy(caption_set)

        captions = []

        for lang in caption_set.get_languages():
            captions.append(self._recreate_lang(caption_set.get_captions(lang)))

        return "".join(captions)

    def _microtoframes(self, micro, fps=25.0):
        return int(micro * fps / (10**6))

    def _recreate_lang(self, captions):
        sub = ""

        for caption in captions:
            start = self._microtoframes(caption.start)
            end = self._microtoframes(caption.end)
            sub += f"{{{start}}}{{{end}}}"

            new_content = ""
            for node in caption.nodes:
                new_content = self._recreate_line(new_content, node)

            # Eliminate excessive line breaks
            new_content = new_content.strip() + "\n"
            while "\n\n" in new_content:
                new_content = new_content.replace("\n\n", "\n")
            # Break unnecessary on last line
            while "|\n" in new_content:
                new_content = new_content.replace("|\n", "\n")

            sub += new_content

        return sub

    def _recreate_line(self, sub, line):
        if line.type_ == CaptionNode.TEXT:
            return sub + line.content
        elif line.type_ == CaptionNode.BREAK:
            return sub + "|"
        else:
            return sub



# We thought about making pycaption.base objects immutable. This would be nice
# in a lot of cases, but since the transformations on them could be quite
# complex, the deepcopy method is good enough sometimes.
from copy import deepcopy
from xml.sax.saxutils import escape

from bs4 import BeautifulSoup

from ..base import BaseWriter, CaptionNode, merge_concurrent_captions
from .base import DFXP_DEFAULT_REGION, DFXPWriter

LEGACY_DFXP_BASE_MARKUP = """
<tt xmlns="http://www.w3.org/ns/ttml"
    xmlns:tts="http://www.w3.org/ns/ttml#styling">
    <head>
        <styling/>
        <layout/>
    </head>
    <body/>
</tt>
"""

LEGACY_DFXP_DEFAULT_STYLE = {
    "color": "white",
    "font-family": "monospace",
    "font-size": "1c",
}

LEGACY_DFXP_DEFAULT_STYLE_ID = "default"
LEGACY_DFXP_DEFAULT_REGION_ID = "bottom"

LEGACY_DFXP_DEFAULT_REGION = {"text-align": "center", "display-align": "after"}


class SinglePositioningDFXPWriter(DFXPWriter):
    """
    A dfxp writer, that ignores all positioning, using a single provided value
    """

    def __init__(self, default_positioning=DFXP_DEFAULT_REGION, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.default_positioning = default_positioning

    def write(self, captions_set, force=""):
        """Writes a DFXP file using the positioning provided in the initializer

        :type captions_set: pycaption.base.CaptionSet
        :param force: only write this language, if available in the CaptionSet
        :rtype: str
        """
        captions_set = self._create_single_positioning_caption_set(
            captions_set, self.default_positioning
        )

        return super().write(captions_set, force)  # noqa

    @staticmethod
    def _create_single_positioning_caption_set(caption_set, positioning):
        """Return a caption where all the positioning information was
        replaced from positioning

        :type caption_set: pycaption.base.CaptionSet
        :rtype: pycaption.base.CaptionSet
        """
        # If SinglePositioningDFXPWriter would modify the state of the caption
        # set, any writer using the same caption_set thereafter would be
        # affected. At the moment we know we don't use any other writers, but
        # this is important and mustn't be neglected
        caption_set = deepcopy(caption_set)
        caption_set = merge_concurrent_captions(caption_set)
        caption_set.layout_info = positioning

        for lang in caption_set.get_languages():
            caption_set.set_layout_info(lang, positioning)

            caption_list = caption_set.get_captions(lang)
            for caption in caption_list:
                caption.layout_info = positioning

                for node in caption.nodes:
                    if hasattr(node, "layout_info"):
                        node.layout_info = positioning

        for _, style in caption_set.get_styles():
            if "text-align" in style:
                style.pop("text-align")

        return caption_set


class LegacyDFXPWriter(BaseWriter):
    """Ported the legacy DFXPWriter from 0.4.5"""

    def __init__(self, *args, **kw):
        self.p_style = False
        self.open_span = False

    def write(self, caption_set, force=""):
        caption_set = deepcopy(caption_set)
        caption_set = merge_concurrent_captions(caption_set)

        dfxp = BeautifulSoup(LEGACY_DFXP_BASE_MARKUP, "lxml-xml")
        dfxp.find("tt")["xml:lang"] = "en"

        for style_id, style in caption_set.get_styles():
            if style != {}:
                dfxp = self._recreate_styling_tag(style_id, style, dfxp)
        if not caption_set.get_styles():
            dfxp = self._recreate_styling_tag(
                LEGACY_DFXP_DEFAULT_STYLE_ID, LEGACY_DFXP_DEFAULT_STYLE, dfxp
            )

        # XXX For now we will always use this default region. In the future if
        # regions are provided, they will be kept
        dfxp = self._recreate_region_tag(
            LEGACY_DFXP_DEFAULT_REGION_ID, LEGACY_DFXP_DEFAULT_REGION, dfxp
        )

        body = dfxp.find("body")

        if force:
            langs = [self._force_language(force, caption_set.get_languages())]
        else:
            langs = caption_set.get_languages()

        for lang in langs:
            div = dfxp.new_tag("div")
            div["xml:lang"] = lang

            for caption in caption_set.get_captions(lang):
                if caption.style:
                    caption_style = caption.style
                    caption_style.update({"region": LEGACY_DFXP_DEFAULT_REGION_ID})
                else:
                    caption_style = {
                        "class": LEGACY_DFXP_DEFAULT_STYLE_ID,
                        "region": LEGACY_DFXP_DEFAULT_REGION_ID,
                    }
                p = self._recreate_p_tag(caption, caption_style, dfxp)
                div.append(p)

            body.append(div)

        caption_content = dfxp.prettify(formatter=None)
        return caption_content

    # force the DFXP to only have one language, trying to match on "force"
    def _force_language(self, force, langs):
        for lang in langs:
            if force == lang:
                return lang

        return langs[-1]

    def _recreate_region_tag(self, region_id, styling, dfxp):
        dfxp_region = dfxp.new_tag("region")
        dfxp_region.attrs.update({"xml:id": region_id})

        attributes = self._recreate_style(styling, dfxp)
        dfxp_region.attrs.update(attributes)

        new_tag = dfxp.new_tag("region")
        new_tag.attrs.update({"xml:id": region_id})
        if dfxp_region != new_tag:
            dfxp.find("layout").append(dfxp_region)
        return dfxp

    def _recreate_styling_tag(self, style, content, dfxp):
        dfxp_style = dfxp.new_tag("style")
        dfxp_style.attrs.update({"xml:id": style})

        attributes = self._recreate_style(content, dfxp)
        dfxp_style.attrs.update(attributes)

        new_tag = dfxp.new_tag("style")
        new_tag.attrs.update({"xml:id": style})
        if dfxp_style != new_tag:
            dfxp.find("styling").append(dfxp_style)

        return dfxp

    def _recreate_p_tag(self, caption, caption_style, dfxp):
        start = caption.format_start()
        end = caption.format_end()
        p = dfxp.new_tag("p", begin=start, end=end)
        p.string = self._recreate_text(caption, dfxp)

        if dfxp.find("style", {"xml:id": "p"}):
            p["style"] = "p"

        p.attrs.update(self._recreate_style(caption_style, dfxp))

        return p

    def _recreate_text(self, caption, dfxp):
        line = ""

        for node in caption.nodes:
            if node.type_ == CaptionNode.TEXT:
                line += escape(node.content) + " "

            elif node.type_ == CaptionNode.BREAK:
                line = line.rstrip() + "<br/>\n    "

            elif node.type_ == CaptionNode.STYLE:
                line = self._recreate_span(line, node, dfxp)

        return line.rstrip()

    def _recreate_span(self, line, node, dfxp):
        if node.start:
            styles = ""

            content_with_style = self._recreate_style(node.content, dfxp)
            for style, value in list(content_with_style.items()):
                styles += f' {style}="{value}"'

            if styles:
                if self.open_span:
                    line = line.rstrip() + "</span> "
                line += f"<span{styles}>"
                self.open_span = True

        elif self.open_span:
            line = line.rstrip() + "</span> "
            self.open_span = False

        return line

    def _recreate_style(self, content, dfxp):
        dfxp_style = {}

        if "region" in content:
            if dfxp.find("region", {"xml:id": content["region"]}):
                dfxp_style["region"] = content["region"]
        if "class" in content:
            if dfxp.find("style", {"xml:id": content["class"]}):
                dfxp_style["style"] = content["class"]
        if "text-align" in content:
            dfxp_style["tts:textAlign"] = content["text-align"]
        if "italics" in content:
            dfxp_style["tts:fontStyle"] = "italic"
        if "font-family" in content:
            dfxp_style["tts:fontFamily"] = content["font-family"]
        if "font-size" in content:
            dfxp_style["tts:fontSize"] = content["font-size"]
        if "color" in content:
            dfxp_style["tts:color"] = content["color"]
        if "display-align" in content:
            dfxp_style["tts:displayAlign"] = content["display-align"]

        return dfxp_style
